require 'dp_stm_map'
require 'timeout'

# RSpec.configure do |c|
#   c.around(:each) do |example|
#     Timeout::timeout(2) {
#       example.run
#     }
#   end
# end

module DpStmMap

  shared_examples_for "StmMap" do


    describe :on_atomic do
      it "should be executed after commiting atomic changes" do
        executed=false
        subject.on_atomic do |changes|          
          executed=true

        end
        subject.atomic do |map|          
          map['x']='y'          
        end
        executed.should be true
      end


      context "when listener raises an error" do
        it "should still commit atomic changes" do
          subject.on_atomic do
            raise "listener failed"
          end
          subject.atomic do |map|
            map['x']='y'
          end

          subject.atomic_read do |map|
            map['x'].should == 'y'
          end
        end
        it "should execute all listeners" do
          executed=false
          subject.on_atomic do
            raise "listener failed"
          end
          subject.on_atomic do |changes|
            executed=true
          end
          subject.atomic do |map|
            map['x']='y'
          end
          executed.should be true
      
        end
      end
    end

    describe :validate_atomic do
      it "should be executed before commiting atomic changes" do
        executed=false
        subject.validate_atomic do |changes|
          executed=true
        end
        subject.atomic do |map|
          map['x']='y'
        end
        executed.should be true
      end

      it "should roll back transaction when validator raises exception" do
        subject.validate_atomic do |changes|
          raise "validation error"
        end

        expect do
          subject.atomic do |map|
            map['x']='y'
          end
        end.to raise_error "validation error"

        subject.atomic_read do |map|
          map.should_not have_key 'x'
        end

      end


      it "should contain changes of creation of new value" do
        subject.validate_atomic do |changes|
          changes.should == {'x' => [nil,'y']}
        end
        subject.atomic do |map|
          map['x']='y'
        end
      end

      it "should contain change to new value" do
        subject.atomic do |map|
          map['x']='y'
        end
        subject.validate_atomic do |changes|
          changes.should == {'x' => ['y','z']}
        end
        subject.atomic do |map|
          map['x']='z'
        end
      end

      it "should contain deleting existing value" do
        subject.atomic do |map|
          map['x']='y'
        end
        subject.validate_atomic do |changes|
          changes.should == {'x' => ['y',nil]}
        end
        subject.atomic do |map|
          map.delete 'x'
        end
      end

    end


    describe :atomic_read do
      it "should return evaluation result" do
        subject.atomic_read{ |map| "test"}.should == 'test'
      end

      it "should propagate raised error" do
        expect{subject.atomic_read{|map| raise "some error"}}.to raise_error "some error"
      end

      describe :[] do
        it "should return nil for unset key" do
          subject.atomic_read do |map|
            map['x'].should == nil
          end
        end

        it "should return value for set key" do
          subject.atomic do |map|
            map['x']='y'
          end
          subject.atomic_read do |map|
            map['x'].should == 'y'
          end
        end

        it "should not create new key when looking up unset key" do
          subject.atomic_read do |map|
            map['x']
            map.should_not have_key 'x'
          end
        end

      end


      describe :has_key? do
        context "when value is previosly set" do
          before do
            subject.atomic do |map|
              map['x']='y'
            end
          end
          it "should return true" do
            subject.atomic_read do |map|
              map.has_key?('x').should be_true
            end
          end
        end

        context "when value is not set" do
          it "should return true" do
            subject.atomic_read do |map|
              map.has_key?('x').should be_false
            end
          end
        end

      end
    end

    describe :atomic do

      it "should return evaluation result" do
        subject.atomic{ |map| "test"}.should == 'test'
      end

      it "should roll back changes when error is raised" do
        expect{subject.atomic{|map| map['x'] = 'y'; raise "some error"}}.to raise_error "some error"
        subject.atomic do |map|
          map.should_not have_key('x')
        end
      end

      describe :has_key? do
        context "when value is previosly set" do
          before do
            subject.atomic do |map|
              map['x']='y'
            end
          end
          it "should return true" do
            subject.atomic do |map|
              map.has_key?('x').should be_true
            end
          end
        end

        context "when value is not set" do
          it "should return true" do
            subject.atomic do |map|
              map.has_key?('x').should be_false
            end
          end
        end

        context "when value is set within transaction" do
          it "should return true" do
            subject.atomic do |map|
              map['x']='y'
              map.has_key?('x').should be_true
            end
          end
        end

        context "when value has been deleted within the transaction" do
          before do
            subject.atomic do |map|
              map['x']='y'
            end
          end
          it "should return false" do
            subject.atomic do |map|
              map.delete 'x'
              map.has_key?('x').should be_false
            end
          end
        end


      end



      describe :delete do

        context "when value does not exist" do
          it "should make no difference" do
            subject.atomic do |map|
              map.delete 'x'
              map.has_key?('x').should be_false
            end
          end
        end

        context "when value exists in the global context" do
          before do
            subject.atomic do |map|
              map['x']='y'
            end
          end
          it "should delete value for the same transaction scope" do
            subject.atomic do |map|
              map.delete 'x'
              map.has_key?('x').should be_false
            end
          end
          it "should delete value for the next transaction scope" do
            subject.atomic do |map|
              map.delete 'x'
            end
            subject.atomic do |map|
              map.has_key?('x').should be_false
            end
          end
        end
      end

      describe :[]= do
        it "should store new value" do
          subject.atomic do |map|
            map['x']='y'
          end
          subject.atomic_read do |map|
            map['x'].should == 'y'
          end
        end

        it "should change existing value" do
          subject.atomic do |map|
            map['x']='y'
          end
          subject.atomic do |map|
            map['x']='z'
          end
          subject.atomic_read do |map|
            map['x'].should == 'z'
          end
        end

        context "when new value is nil" do
          before do
            subject.atomic {|map| map['x']='y'}
          end

          it "should delete the key in the same transaction scope" do
            subject.atomic do |map|
              map['x']=nil
              map.should_not have_key 'x'
            end
          end

          it "should delete the key in the global scope after transaction" do
            subject.atomic do |map|
              map['x']=nil
            end
            subject.atomic do |map|
              map.should_not have_key 'x'
            end
          end

        end


        context "when value has been previosly deleted in the same transaction scope" do
          it "should store the new value visible within the transaction scope" do
            subject.atomic do |map|
              map['x'] = 'y'
              map.delete 'x'
              map['x'] = 'z'
              map['x'].should == 'z'
            end
          end

          it "should store the new value visible within scope of next transaction" do
            subject.atomic do |map|
              map['x'] = 'y'
              map.delete 'x'
              map['x'] = 'z'
            end
            subject.atomic do |map|
              map['x'].should == 'z'
            end
          end
        end


      end

      describe :[] do


        it "should record reading values" do
          subject.atomic do |map|
            map['x']='y'            
          end

          subject.validate_atomic do |changes|
            changes.should == {'x' => ['y','y']}
          end

          subject.atomic do |map|
            map['x']
          end

        end

        it "should record changing values" do
          subject.atomic do |map|
            map['x']='y'            
          end

          subject.validate_atomic do |changes|
            changes.should == {'x' => ['y','z']}
          end

          subject.atomic do |map|
            map['x']='z'
          end

        end

        it "should return nil for unset key" do
          subject.atomic do |map|
            map['x'].should == nil
          end
        end

        it "should not create new key when looking up unset key" do
          subject.atomic do |map|
            map['x']
            map.should_not have_key 'x'
          end
        end

        context "when value is not set in local atomic context" do
          before do
            subject.atomic do |map|
              map['x']='y'
            end
          end
          it "should return value from global state" do
            subject.atomic do |map|
              map['x'].should == 'y'
            end
          end
        end

        context "when value is set in local atomic context" do
          it "should return value from the local state" do
            subject.atomic do |map|
              map['x']= 'z'
              map['x'].should == 'z'
            end
          end
        end


      end

    end



  end

end
